(* ::Package:: *)

(* ::Section:: *)
(*Bigquery Package*)


(* ::Subsection:: *)
(*About*)


(* By Rodrigo Murta - murta@rodrigomurta.com   *)
(* Last edition: 2021-01-27                    *)
(* JWT Authentication by Pedro Cabral          *)


(* ::Subsection:: *)
(*Declarations*)


BeginPackage["Bigquery`"];


bigqueryGetToken::usage="bigqueryGetToken[]";
bigqueryExecute::usage="bigqueryExecute[query_String]";


(* ::Section:: *)
(*Functions*)


(* ::Subsection:: *)
(*Begin Package*)


Begin["`Private`"]


(* ::Subsection:: *)
(*JWT Authentication*)


$keyJson = "bigquery-key.json";
$thisFileDirectory = FileNameDrop@If[$InputFileName == "", NotebookFileName[], $InputFileName];
$jsonJWTPath = FileNameJoin@{$thisFileDirectory, "bigquery-key.json"};
$projectId = Import[$jsonJWTPath, "RawJSON"]["project_id"];


fixBase64::invarg = "The argument `` is not a valid string.";
encodeBase64::invarg = "The argument `` is not a valid string.";


fixBase64[str_] := Module[{},

	(* If the string argument is not a valid string, return an error message. *)
	If[!StringQ[str],
		Return[Message[fixBase64::invarg, str]];
	];
	
	(* If the string is empty, return an empty string. *)
	(* Removing this can lead to catastrophic errors. *)
	If[SameQ[str, ""],
		Return[""];
	];
	
	(* Remove all "=", and replace "+" with "-" and "/" with "_". *)
	StringReplace[StringSplit[str, "="][[1]], {"+" -> "-", "/" -> "_"}]
];


(* Note: Using ExportString[str, "Base64"] is extremely inefficient. *)
encodeBase64[str_] := Module[{},

	(* If the string argument is not a valid string, return an error message. *)
	If[!StringQ[str],
		Return[Message[encodeBase64::invarg, str]];
	];
	
	(* Return the fixed base-64 string. *)
	fixBase64[Developer`EncodeBase64[str]]
];


bigqueryGetToken::invjwt="``";
bigqueryGetToken::invpath="Argument `` is an invalid path.";


(* This function builds and sends a JWT request to "oauth2.googleapis.com" for retrieval of the Token. *)
bigqueryGetToken[path_] := Module[{jsonkey, jhead, jreq, pkey, encrypted, url, res},
	
	(* If the credentials path does not exist, return an error. *)
	If[!FileExistsQ[path],
		Return[Message[bigqueryGetToken::invpath, path]]
	];
	
    (* Open JSON File. *)
    jsonkey = Import[path, "RawJSON"];

    (* Build "Header" and "Payload" JSON. *)
    jhead = encodeBase64@ExportString[<|"alg" -> "RS256", "typ" -> "JWT"|>, "JSON", "Compact" -> True];
    jreq = encodeBase64@StringReplace[
			ExportString[
				<|"iss" -> StringTrim[jsonkey[["client_email"]]]
				, "scope" -> "https://www.googleapis.com/auth/bigquery"
				, "aud" -> "https://oauth2.googleapis.com/token"
				, "iat" -> UnixTime[]
				, "exp" -> UnixTime[] + 3600
				|>
				, "JSON"
				, "Compact" -> True
			]
			, "\\" -> ""
		];

    (* Step 4: Open PrivateKey and Encrypt "Header" and "Payload" together. *)
    pkey = First@ImportString[StringTrim[jsonkey[["private_key"]]], "PEM"];
    encrypted = fixBase64@BaseEncode[GenerateDigitalSignature[jhead ~~ "." ~~ jreq, pkey][[1, -1]]];

    (* Step 5: Build the URL. *)
    url = HTTPRequest[
			<|"Scheme" -> "https"
				, "Domain" -> "oauth2.googleapis.com"
				, "Path" -> "token"
				, "Method" -> "POST"
				, "ContentType" -> "application/x-www-form-urlencoded"
				, "Query" -> {
						"grant_type" -> "urn:ietf:params:oauth:grant-type:jwt-bearer"
						, "assertion" -> jhead <> "." <> jreq <> "." <> encrypted}
			|>
			, VerifySecurityCertificates -> True
		];

    (* Step 6: Execute the POST Request. *)
    res = URLExecute[url];

    (* Handle JWT Errors. *)
    If[KeyExistsQ[res, "error"], Return@Message[bigqueryGetToken::invjwt, res[[2, 2]]]];

	(* Return the token. *)
	Return@res[[1, 2]];
]
(*bigqueryGetToken[$jsonJWTPath]*)


bigqueryGetToken[]:= bigqueryGetToken[$jsonJWTPath]
(*bigqueryGetToken[]*)


(* ::Subsection:: *)
(*Conversor*)


(* Attributes definition. *)
SetAttributes[typeConverter, Listable];


(* Converts an integer with FromDigits[]. *)
typeConverter["INTEGER", value_String] := FromDigits[value];


(* Converts a float with Internal`StringToDouble[] or ImportString[]. *)
typeConverter["FLOAT", value_String] := Internal`StringToDouble[value] /. $Failed["Bignum"] :> First @ ImportString[value, "List"];


(* Converts a numeric value with Internal`StringToDouble[] or ImportString[]. *)
typeConverter["NUMERIC", value_String] := Internal`StringToDouble[value] /. $Failed["Bignum"] :> First @ ImportString[value, "List"];


(* Converts a boolean with simple substitution. *)
typeConverter["BOOLEAN", value_String] := value /. {"true" -> True, "false" -> False};


(* Converts a timestamp using FromUnixTime[] and DateString[]. *)
typeConverter["TIMESTAMP", value_String] := DateString[FromUnixTime[Internal`StringToDouble[value]], {"Year", "-", "Month", "-", "Day", " ", "Hour", ":", "Minute", ":", "SecondExact"}];


(* These data types can be altered to any particular needs. *)
typeConverter["DATE", value_String] := value;
typeConverter["TIME", value_String] := value;
typeConverter["STRING", value_String] := value;
typeConverter["BYTES", value_String] := value;
typeConverter["GEOGRAPHY", value_String] := value;
typeConverter["RECORD", value_String] := value;


(* For other types, no processing should be done. *)
typeConverter[_, value_] := value;


(* ::Subsection:: *)
(*Bigquery Execute*)


(*https://stackoverflow.com/questions/45108038/how-can-i-use-the-bigquery-rest-api-from-the-command-line*)
(*https://stackoverflow.com/a/53529534/2048848*)
bigqueryExecute // ClearAll
Options[bigqueryExecute] = {"token" :> bigqueryGetToken[], "timeout" -> 30000, "maxResults"-> 50000};
bigqueryExecute[query_String, opts:OptionsPattern[{bigqueryExecute}]]:= Module[{url, token, nextPageToken, nextJobId, r, nextRecordsUrl = Null, request, fields, rawValues, types},
	
	token = OptionValue["token"];
	
	(* This URL will be used for querying. *)
	url = "https://bigquery.googleapis.com/bigquery/v2/projects/"<>$projectId<>"/queries";
	
	(* Build the request. *)
	request = HTTPRequest[url,
				<|Method -> "Post"
				, "Headers" -> {"Authorization"-> "Bearer "<>token}
				, "Body" ->  (<|
							  "kind"-> "bigquery#queryRequest"
							,"useLegacySql" -> False
							,"query" -> query
							,"timeoutMs" -> OptionValue["timeout"]
							,"maxResults" -> OptionValue["maxResults"]
							|>//ExportString[#, "RawJSON"]&)
				|>];
	
	(* Read the URL and get the response from JSON. *)
	r = FromCharacterCode@URLRead[request, "BodyBytes"]//ImportString[#, "RawJSON"]&;
	
	(* These two lines get the NextPageToken and NextJobId for the next part of the querying process. *)
	(* Echo@KeyDrop[res, {"schema", "rows"}]; *)
	nextPageToken = Lookup[r,"pageToken", None];
	nextJobId = r[["jobReference", "jobId"]];
	
	(* Check if the job is complete. *)
	If[r["jobComplete"] === False, Return[r]];
	If[KeyExistsQ[r, "error"], Return[r]];
	If[r["totalRows"] === "0", Return[<||>]];
	
	(* Do some preprocessing to the data. *)
	fields = r[["schema", "fields", All, "name"]];
	types = r[["schema", "fields", All, "type"]];
	rawValues = Transpose@r[["rows",All,"f",All,"v"]];
	r = Inner[typeConverter,types,rawValues,List];
	r = AssociationThread[fields -> #]&/@r;
	
	(* If there's a NextPageToken, execute a new BigQuery query with the NextPageToken. *)
	If[nextPageToken =!= None,
			Echo["Next Page"];
			r = Join[r, bigqueryExecute[{nextJobId, nextPageToken}, opts]]
	];
	
	(* Return the response. *)
	r
]
(*r = bigqueryExecute["select 1 as test"]*)


(* This function continues a BigQuery query with a JobID and a NextPageToken. *)
(*https://cloud.google.com/bigquery/docs/reference/rest/v2/jobs/getQueryResults*)
bigqueryExecute[{jobId_, pageToken_}, opts:OptionsPattern[{bigqueryExecute}]]:= Module[{url, token, nextPageToken, nextJobId, r, request, fields, rawValues, types},
	
	token = OptionValue["token"];
	
	(* This URL will be used for querying. *)
	url = "https://bigquery.googleapis.com/bigquery/v2/projects/"<>$projectId<>"/queries/";
	url = URLBuild[{url,jobId},{"pageToken" -> pageToken, "timeoutMs" -> OptionValue["timeout"], "maxResults" -> OptionValue["maxResults"]}];
	
	(* Build the request. *)
	request = HTTPRequest[url,
				<|Method -> "Get"
				, "Headers" -> {"Authorization"-> "Bearer "<>token}
				|>];
	
	(* Read the URL and get the response from JSON. *)
	r = FromCharacterCode@URLRead[request, "BodyBytes"]//ImportString[#, "RawJSON"]&;
	
	(* These two lines get the NextPageToken and NextJobId for the next part of the querying process. *)
	nextPageToken = Lookup[r,"pageToken", None];
	nextJobId = r[["jobReference", "jobId"]];

	(* Check if the job is complete. *)
	If[r["jobComplete"] === False, Return[r]];
	If[KeyExistsQ[r, "error"], Return[r]];
	If[r["totalRows"] === "0", Return[<||>]];
	
	(* Do some preprocessing to the data. *)
	fields = r[["schema", "fields", All, "name"]];
	types = r[["schema", "fields", All, "type"]];
	rawValues = Transpose@r[["rows",All,"f",All,"v"]];
	r = Inner[typeConverter,types,rawValues,List];
	r = AssociationThread[fields -> #]&/@r;
	
	(* If there's a NextPageToken, execute a new BigQuery query with the NextPageToken. *)
	If[nextPageToken =!= None,
			Echo["Next Page"];
			r = Join[r, bigqueryExecute[{nextJobId, nextPageToken}, opts]]
	];

	(* Return the response. *)
	r
]
(*bigqueryExecute[{"job_6HruT0ERH-", "BFI47FQMOMAQAAAS"}]*)


(* ::Subsection:: *)
(*End Package*)


End[]
EndPackage[]
